/*! @file
 * @brief
 * contains stuff related to statistical information about the progress of factorization
 */


//! data structures and objects related to statistical information about the progress of factorization
namespace statistical_data
{
 int DynamicFactorRating[7] = {0,0,0,0,0,0,0};
  /* for statistical evaluation on how the dynamic factorbase
     is responsible for the finding of relations:
      [0]-> no dynamic factor is involved,
      [1]-> one dynamic factor is involved
      [2]-> two dynamic factors are involved
      [3]-> three dynamic factors are involved
      [4]-> four dynamic factors are involved
      [5]-> 5 dynamic factors are involved
      [6]-> more than 5 dynamic factors are involved
   */

 int directly_sieved_DynamicFactors = 0;
 int directly_sieved_SpecialFactors = 0;

#ifndef IS_CLIENT
 // the following statistical data can only be collected during
 // a detailed analysis of the relations. Therefore these data
 // is restricted to servers and stand-alone versions.
 int Special_hit = 0;
 int Special_to_dynamic_Factor_hit = 0;
 int gefundene_DLP_Zyklen = 0;
#endif

#ifdef IS_SERVER
 // only for a server: collect statistical data about clients

 class CClientStats : protected CmpqsFactortypes
 {
  private:
   static CMutex Mutex;
   static unsigned int overall_counter[Factortype_size];
   unsigned int counter[Factortype_size];
   time_t first_connect, timestamp;
  protected:
   static void lock() { Mutex.lock(); }
   static void unlock() { Mutex.unlock(); }
  public:

   CClientStats() : counter(), first_connect(time(NULL)), timestamp(0) { };
   ~CClientStats() { };

   void PutTimeStamp(const time_t t = time(NULL))
   {
     lock();
     timestamp=t;
     unlock();
   }

   /// increment the counter for the given type or factor (eg. static,SLP,DLP)
   void increment(const CmpqsFactortypes::Factortype FType)
   {
     lock();
     ++counter[FType]; ++overall_counter[FType];
     unlock();  
   }
  
   const string info() const
   {
     ostringstream os;
     char buf [32];
     lock();
     const double percentage =
      100.0*static_cast<double>(counter[static_prime]+counter[single_large_prime]+counter[double_large_prime])
       /static_cast<double>(overall_counter[static_prime]+overall_counter[single_large_prime]+overall_counter[double_large_prime]);
     ctime_r(&first_connect,buf); // get localtime string for first_connect
     //buf[24]='\0'; // get rid of trailing newline
     buf[16]='\0'; // cut seconds, year and trailing newline
     char* const datestr = &buf[4]; // be less verbose
     os << datestr << " - ";
     ctime_r(&timestamp,buf);
     buf[16]='\0'; // cut seconds, year and trailing newline
     os << datestr << ":"
        << setw(5) << counter[static_prime] << ","
        << setw(7) << counter[single_large_prime] << ","
        << setw(8) << counter[double_large_prime] << "|";
     os.setf(ios::right, ios::adjustfield);
     os.setf(ios::fixed, ios::floatfield);
     os << setw(5) << setprecision(1) << percentage << "%" << flush;
     unlock();
     return os.str();
   }

   const string XML_info() const
   {
     ostringstream os;
     char buf [32];
     lock();
     const double percentage =
      100.0*static_cast<double>(counter[static_prime]+counter[single_large_prime]+counter[double_large_prime])
       /static_cast<double>(overall_counter[static_prime]+overall_counter[single_large_prime]+overall_counter[double_large_prime]);
     ctime_r(&first_connect,buf); // get localtime string for first_connect
     //buf[24]='\0'; // get rid of trailing newline
     buf[16]='\0'; // cut seconds, year and trailing newline
     char* const datestr = &buf[4]; // be less verbose
     os << " FIRSTCONNECT=\"" << datestr << "\"";
     ctime_r(&timestamp,buf);
     buf[16]='\0'; // cut seconds, year and trailing newline
     os << " LASTCONNECT=\"" << datestr << "\""
        << " STATICRELATIONS=\"" << counter[static_prime] << "\""
        << " DYNAMICRELATIONS=\"" << counter[single_large_prime] << "\""
        << " DLPRELATIONS=\"" << counter[double_large_prime] << "\"";
     os.setf(ios::right, ios::adjustfield);
     os.setf(ios::fixed, ios::floatfield);
     os << " PERCENTAGE=\"" << setw(5) << setprecision(1) << percentage
        << "%" << "\"" << flush;
     unlock();
     return os.str();
   }

 };
 CMutex CClientStats::Mutex;
 unsigned int CClientStats::overall_counter[Factortype_size] = {0};

 typedef map<string,CClientStats> TAllClientStats;
 TAllClientStats AllClientStats; // this is a list of all clients ever connected to the server
 CMutex AllClientStats_Mutex;

#endif /* IS_SERVER */

 class CProgressStats : private ForbidAssignment
  {
   private:
    double granularity;
    double next_sample_d;
    int next_sample_at;

    class TSample
     {
      private:
       unsigned int static_relations, SLP_relations, DLP_relations;
       time_t Time;
      public:
       TSample() : static_relations(StaticRelations::Count()),
          SLP_relations(DynamicFactorRelations.size()),
          DLP_relations(SpecialRelations::Count()), Time(time(NULL)) { }

       ~TSample() { }

       void XML(std::ostream &os) const
        {
          char buf [32];
          ctime_r(&Time,buf); // get localtime
          buf[24]='\0'; // get rid of trailing newline
          char* const datestr = &buf[4]; // be less verbose
          os << "<tr>"
             << "<td>" << setw(5) << setprecision(4) << 100.0*static_cast<double>(static_relations)/static_cast<double>(StaticFactorbase::Size()) << "%</td>"
             << "<td>" << static_relations << "</td>"
             << "<td>" << SLP_relations << "</td>"
             << "<td>" << DLP_relations << "</td>"
             << "<td>" << datestr << "</td>"
             << "</tr>";
        }
     };

    typedef std::vector<TSample> TSamples;
    TSamples Samples;

   public:
    explicit CProgressStats(const double _granularity = 0.01)
     : granularity(_granularity), next_sample_d(0.0), next_sample_at(0) { }

    ~CProgressStats() { }

   public:
    const void take_sample()
     {
       if (StaticRelations::Count()<next_sample_at) return;
       Samples.push_back(TSample()); // create & append new sample
       do {
         next_sample_d += granularity;
         next_sample_at = static_cast<int>(StaticFactorbase::Size()*next_sample_d);
        } while (next_sample_at<StaticRelations::Count());
     }

    void XML(std::ostream &os) const
     {
       if (Samples.empty()) return;
       os << "<TABLE CONTENT=\"progress\" >" << endl;
       os << "<tr><td>Rate</td><td>Static Relations</td><td>SLP Relations</td><td>DLP Relations</td><td>Time</td></tr>" << endl;
       for (TSamples::const_iterator pos=Samples.begin(); pos!=Samples.end(); ++pos)
        {
          pos->XML(os); os << endl;
        }
       os << "</TABLE>" << endl;
     }
  };

 CProgressStats ProgressStats;

 time_t StartTimestamp = 0; // Zeitpunkt, zu dem das Sieben gestartet wurde
 time_t PreviousTimestamp = 0; // Zeitpunkt der vorigen Meldung
 int StartFuellung; // Fllung des GLS bei Start des Aussiebens
 struct tsample { time_t Zeitpunkt; int Fuellung; };
 const int samples=1<<4;
 tsample sample[samples]; int i_newest_sample=0;

 void calc_ETA(void)
 {
   // eigentlich wird hier weniger die Restzeit abgeschtzt als
   // vielmehr die durchschnittlich bentigte Zeit zum Sieben einer
   // Relation ermittelt...
   // Fr eine Restzeitabschtzung mte man das Zeitverhalten eigentlich
   // extrapolieren (und zustzlich die Schwankungen zwischen den
   // "Zwischenankunftszeiten" der Relationen ausmitteln).
   // Dies wre aber ein hoffnungsloses Unterfangen, da die Systemlast
   // schwankt und (im Serverbetrieb) die Anzahl der siebenden Clients
   // unbekannt ist (und zudem auch die Fileclients die Ergebnisse
   // verflschen wrden).
   if (StaticRelations::Count()==StartFuellung) return;
   const time_t now=time(NULL);
   const int i_oldest_sample=(i_newest_sample+1)%samples;

   /*
      now.x - old.x = delta.x  // relations
      now.y - old.y = delta.y  // time
      delta.y/delta.x = y per x  -> seconds per static relation

      wanted.x = missing relations = StaticFactorbase::Size() - StaticRelations::Count(); known
      wanted.y -> (estimated) time till arrival; to compute

      wanted.y = (delta.y/delta.x)*wanted.x
      ETA = estimated time of arrival = now.y + wanted.y

   */

   double t=difftime(now,sample[i_oldest_sample].Zeitpunkt);
   t/=static_cast<double>(StaticRelations::Count()-sample[i_oldest_sample].Fuellung);
   double Restzeit=(StaticFactorbase::Size()-StaticRelations::Count())*t;

   {
     // print estimated time of arrival
     char buf[32]; struct tm my_tm;
     const time_t eta = now+static_cast<time_t>(Restzeit);
     localtime_r(&eta,&my_tm); asctime_r(&my_tm,buf);
     buf[24]='\0'; // get rid of trailing newline
     cout_status << "ETA: " << buf << ", ";
   }

   double days = floor(Restzeit/(60*60*24));
   int hours,mins,secs;
   Restzeit=fmod(Restzeit,60*60*24);
   secs=static_cast<int>(Restzeit); mins=secs/60; hours=mins/60;
   mins%=60; secs%=60;

   if (days)
    cout_status << "time left (estimated): " << days << "d " << hours << "h" << endl;
   else
    cout_status << "time left (estimated): " << hours << "h " << mins << "m "
         << secs << "s" << endl;

#if defined(VERBOSE)
   cout_status << "oldest sample: " << sample[i_oldest_sample].Zeitpunkt << ": " << sample[i_oldest_sample].Fuellung << endl;
   cout_status << "newest sample: " << sample[i_newest_sample].Zeitpunkt << ": " << sample[i_newest_sample].Fuellung << endl;
#endif

   // Ringbuffer nur dann aktualisieren, wenn
   // mind. 1 Minute vergangen und Fllung sich seitdem erhht hat
   if (now-sample[i_newest_sample].Zeitpunkt>=60 && StaticRelations::Count()>sample[i_newest_sample].Fuellung)
    {
      i_newest_sample=i_oldest_sample; // wegen Ringbuffer!!
      sample[i_newest_sample].Zeitpunkt=now;
      sample[i_newest_sample].Fuellung=StaticRelations::Count();
    }
 }

 void display_StatusLegend(void)
 {
   cout_status << "Legend of the status information:" << endl;
   cout_status << "#static relations{distribution}/#static factorbase," << endl
               << "#directly sieved dynamic factors (single large primes)/#dynamic-FB," << endl;
 #ifdef IS_CLIENT
   cout_status << "#special factors (double large primes) ->{distribution}" << endl;
 #else
   cout_status << "#special factors (double large primes)" << endl;
 #endif
 #ifndef USE_NETWORK
   // Sieboffset macht nur bei standalone-Version Sinn
   cout_status << " [sievepos]" << endl;
 #endif
   cout_status << " -> % progress" << endl;

   if (!StartTimestamp)
    {
     StartTimestamp=time(NULL); StartFuellung=StaticRelations::Count(); // Initialisierung fr Restzeitberechnung
     PreviousTimestamp=StartTimestamp; // Zeitpunkt dieser Ausgabe merken
     for (int i=0; i<samples; ++i) { sample[i].Zeitpunkt=StartTimestamp; sample[i].Fuellung=StaticRelations::Count(); }
    }
 }

 bool StatusReport(const bool force_now = false)
 {
   /*
      Statusmeldung ausgeben, wenn seit voriger Statusmeldung/Statuslegende
      mindestens 15 Sekunden vergangen sind.
      Der Funktionswert ist true, falls eine Ausgabe erfolgte, ansonsten false
    */ 

   if (!force_now && difftime(time(NULL),PreviousTimestamp)<15.0) return false;
   // aktuellen Stand ausgeben
   
   ProgressStats.take_sample();

   cout_status << "Status for " << n << endl;

 #ifdef IS_SERVER
   {
    // print a sheet with all clients that ever have connected to this server
    AllClientStats_Mutex.lock();
    for (TAllClientStats::iterator pos=AllClientStats.begin(); pos!=AllClientStats.end(); ++pos)
      cout_status << (pos->first+"                    ").substr(0,20) << ":" << pos->second.info() << endl;
    AllClientStats_Mutex.unlock();
   }
 #endif

   // overall statistics
   cout_status << StaticRelations::Count()
        << "{" << DynamicFactorRating[0]
        << "," << DynamicFactorRating[1]
        << "," << DynamicFactorRating[2]
        << "," << DynamicFactorRating[3]
        << "," << DynamicFactorRating[4]
        << "," << DynamicFactorRating[5]
        << "," << DynamicFactorRating[6]
        << "}/" << StaticFactorbase::Size() << ","
        << directly_sieved_DynamicFactors << "/"
        << DynamicFactorRelations.size();

 #ifdef IS_SERVER
   cout_status << "{A:" << DynamicFactorRelations.size_active()
      << ",P:" << DynamicFactorRelations.size_passive()
      << "}";
 #endif

   cout_status << ",";
 #ifdef IS_CLIENT
   cout_status << directly_sieved_SpecialFactors;
 #else
   cout_status
        << SpecialRelations::Count()
           // BUG/remark: SpecialFactorRelations is not protected by a
           // mutex, we actually can get wrong results, if we read
           // SpecialFactorRelations.size() while another thread is
           // modifying the set. But since we do read-only access, this
           // should not cause any harm. (this actually happened once while
           // Zyklensuche was active)
        << "->{" << Special_to_dynamic_Factor_hit << "/" << Special_hit
        << "/" << gefundene_DLP_Zyklen << "}";
 #endif
 #ifndef USE_NETWORK
   // Sieboffset macht nur bei standalone-Version Sinn
   cout_status << " [" << SieveOffset << "]";
 #endif
   cout_status.setf(ios::left, ios::adjustfield);
   cout_status << " -> ca. " << setw(5) << setprecision(4)
        << 100.0*StaticRelations::Count()/StaticFactorbase::Size()
        << "%" << endl;
   cout_status << "mpqs interval ";
   Polynom.save(cout_status); // mpqs-Polynomdaten ausgeben
   PreviousTimestamp=time(NULL); // Zeitpunkt dieser Ausgabe merken
 #ifndef IS_CLIENT
   calc_ETA();
 #endif
   return true;
 }

 void XML_StatusReport(ostream &os)
 {
   {
     time_t t = time(NULL);
     char buf [32];
     ctime_r(&t,buf); // get localtime
     buf[24]='\0'; // get rid of trailing newline
     char* const datestr = &buf[4]; // be less verbose
     os << "<QSIEVESTATISTICS TIME=\"" << datestr
        << "\" VERSION=\"" << VERSION << "\" >" << endl;
   }

   os << "<NUMBERSECTION>" << endl;
   if (FoundFactors.regarding!="") os << "<ITEM CONTENT=\"input: \">" << FoundFactors.regarding << "</ITEM>" << endl;
   if (mpz_cmp_ui(n,1)!=0) os << "<ITEM CONTENT=\"remaining: \">" << n << "</ITEM>" << endl;
   os << "</NUMBERSECTION>" << endl;

   if (!FoundFactors.empty())
    {
//  FIXME! TFoundFactors needs to be wrapped (not threadsafe!)
      os << "<TABLE CONTENT=\"FACTORS\" >";
       for (TFoundFactors::iterator pos=FoundFactors.begin(); pos!=FoundFactors.end(); ++pos)
        os << "<FACTOR>" << (*pos) << "</FACTOR>";
      os << "</TABLE>" << endl;
    }

   if (ecm_curves_processed)
    {
      os << "<TABLE CONTENT=\"Elliptic Curves\" >"
         << "<tr><td>ecm status</td><td>" << ecm_curves_processed << " / " << elcu_Kurven << "</td></tr>" << endl
         << "<tr><td>Phase 1</td><td>" << elcu_Phase1 << "</td></tr>" << endl
         << "<tr><td>Phase 2</td><td>" << elcu_Phase2 << "</td></tr>" << endl
         << "</TABLE>" << endl;
    }

   ProgressStats.XML(os); // a table with progress statistics
 #ifdef IS_SERVER
   {
    // print a sheet with all clients that ever have connected to this server
    AllClientStats_Mutex.lock();
    if (!AllClientStats.empty())
     {
       os << "<TABLE CONTENT=\"NETCLIENTS\" >" << endl;
       for (TAllClientStats::iterator pos=AllClientStats.begin(); pos!=AllClientStats.end(); ++pos)
        os << "<NETCLIENT "
           << " HOSTNAME=\"" << pos->first << "\""
           << pos->second.XML_info()
           << " />" << endl;
       os << "</TABLE>" << endl;
     }
    AllClientStats_Mutex.unlock();
   }
 #endif

   // overall statistics
   os << "<STATUSINFO>" << endl;
   os << "<RELFOUND>" << StaticRelations::Count()
      << "<RELDISTRIBUTION>"
      << DynamicFactorRating[0] << ", "
      << DynamicFactorRating[1] << ", "
      << DynamicFactorRating[2] << ", "
      << DynamicFactorRating[3] << ", "
      << DynamicFactorRating[4] << ", "
      << DynamicFactorRating[5] << ", "
      << DynamicFactorRating[6] << " </RELDISTRIBUTION>"
      << "</RELFOUND>" << endl
      << "<RELTOTAL>" << StaticFactorbase::Size() << "</RELTOTAL>" << endl
      << "<SLPFOUND>" << directly_sieved_DynamicFactors << "</SLPFOUND>"
      << "<SLPTOTAL>" << DynamicFactorRelations.size();

 #ifdef IS_SERVER
   os << "<RELDISTRIBUTION>" << "A:" << DynamicFactorRelations.size_active()
      << ", P:" << DynamicFactorRelations.size_passive()
      << " </RELDISTRIBUTION>";
 #endif

   os << "</SLPTOTAL>";

   os << "<DLPTOTAL>" << SpecialRelations::Count() << "</DLPTOTAL>" << endl;
           // BUG/remark: SpecialFactorRelations is not protected by a
           // mutex, we actually can get wrong results, if we read
           // SpecialFactorRelations.size() while another thread is
           // modifying the set. But since we do read-only access, this
           // should not cause any harm. (this actually happened once while
           // Zyklensuche was active)

 #ifndef IS_CLIENT
   os << "<DLP2SLP>" << Special_to_dynamic_Factor_hit << "</DLP2SLP>"
      << "<DLP2STATIC>" << Special_hit << "</DLP2STATIC>"
      << "<DLP2CYCLE>" << gefundene_DLP_Zyklen << "</DLP2CYCLE>" << endl;
 #endif

   os.setf(ios::left, ios::adjustfield);
   os << "<PERCENTAGEDONE>" << setw(5) << setprecision(4)
        << 100.0*StaticRelations::Count()/StaticFactorbase::Size()
        << "%" << "</PERCENTAGEDONE>" << endl;
   os << "</STATUSINFO>" << endl;
   os << "</QSIEVESTATISTICS>" << endl;
 }



} // end of namespace statistical_data
using statistical_data::StatusReport;
using statistical_data::display_StatusLegend;
